@@ -1,3 +1,4 @@
+# pylint: disable=C0302
 """
 This module provides SQL query parsing functions
 """
@@ -49,9 +50,12 @@ class Parser:  # pylint: disable=R0902
         self._table_aliases = None
 
         self._with_names = None
+        self._with_queries = None
+        self._with_queries_columns = None
         self._subqueries = None
         self._subqueries_names = None
         self._subqueries_parsers = dict()
+        self._with_parsers = dict()
 
         self._limit_and_offset = None
 
@@ -170,6 +174,7 @@ class Parser:  # pylint: disable=R0902
         columns = UniqueList()
         tables_aliases = self.tables_aliases
         subqueries_names = self.subqueries_names
+        with_names = self.with_names
 
         for token in self.tokens:
             # handle CREATE TABLE queries (#35)
@@ -215,17 +220,28 @@ class Parser:  # pylint: disable=R0902
                             token.previous_token.is_right_parenthesis
                             and token.value in subqueries_names
                         )
+                        and not (
+                            # names of the with queries <name> as (subquery)
+                            token.next_token.normalized == "AS"
+                            and token.value in with_names
+                        )
                         # custom functions - they are followed by the parenthesis
                         # e.g. custom_func(...
                         and not token.next_token.is_left_parenthesis
                     ):
                         column = token.table_prefixed_column(tables_aliases)
-                        column = self._resolve_subquery_alias_to_column(column)
+                        if self._is_with_query_already_resolved(column):
+                            self._add_to_columns_aliases_subsection(
+                                token=token, left_expand=False
+                            )
+                            continue
+                        column = self._resolve_sub_queries(column)
                         self._add_to_columns_with_tables(token, column)
                         self._add_to_columns_subsection(
                             keyword=token.last_keyword_normalized, column=column
                         )
                         columns.extend(column)
+
                 elif (
                     token.last_keyword_normalized == "INTO"
                     and token.previous_token.is_punctuation
@@ -245,7 +261,7 @@ class Parser:  # pylint: disable=R0902
             ):
                 # handle * wildcard in select part, but ignore count(*)
                 column = token.table_prefixed_column(tables_aliases)
-                column = self._resolve_subquery_alias_to_column(column)
+                column = self._resolve_sub_queries(column)
                 self._add_to_columns_with_tables(token, column)
                 self._add_to_columns_subsection(
                     keyword=token.last_keyword_normalized, column=column
@@ -255,15 +271,6 @@ class Parser:  # pylint: disable=R0902
         self._columns = columns
         return self._columns
 
-    @property
-    def columns_without_subqueries(self) -> List:
-        """
-        Returns columns without ones explicitly coming from sub-queries
-        """
-        columns = self.columns
-        subqueries = self.subqueries_names
-        return [column for column in columns if column.split(".")[0] not in subqueries]
-
     @property
     def columns_dict(self) -> Dict[str, List[str]]:
         """
@@ -429,6 +436,12 @@ class Parser:  # pylint: disable=R0902
                     and token.is_create_table_columns_definition
                 ):
                     continue
+                if (
+                    token.normalized == "WITH"
+                    and token.previous_token.is_left_parenthesis
+                    and token.get_nth_previous(2).normalized == "FROM"
+                ):
+                    continue
 
                 if token.next_token.is_dot:
                     pass  # part of the qualified name
@@ -562,6 +575,45 @@ class Parser:  # pylint: disable=R0902
         self._with_names = with_names
         return self._with_names
 
+    @property
+    def with_queries(self) -> Dict[str, str]:
+        """
+        Returns "WITH" subqueries with names
+
+        E.g. WITH tableFromWith AS (SELECT * FROM table3)
+             SELECT "xxxxx" FROM database1.tableFromWith alias
+             LEFT JOIN database2.table2 ON ("tt"."ttt"."fff" = "xx"."xxx")
+        will return {"tableFromWith": "SELECT * FROM table3"}
+        """
+        if self._with_queries is not None:
+            return self._with_queries
+        with_queries = dict()
+        with_queries_columns = dict()
+        for name in self.with_names:
+            token = self.tokens[0].find_nearest_token(
+                name, value_attribute="left_expanded", direction="right"
+            )
+            if token.next_token.is_with_columns_start:
+                with_queries_columns[name] = True
+            else:
+                with_queries_columns[name] = False
+            current_with_query = []
+            with_start = token.find_nearest_token(
+                True, value_attribute="is_with_query_start", direction="right"
+            )
+            with_end = with_start.find_nearest_token(
+                True, value_attribute="is_with_query_end", direction="right"
+            )
+            query_token = with_start.next_token
+            while query_token != with_end:
+                current_with_query.append(query_token)
+                query_token = query_token.next_token
+            with_query_text = "".join([x.stringified_token for x in current_with_query])
+            with_queries[name] = with_query_text
+        self._with_queries = with_queries
+        self._with_queries_columns = with_queries_columns
+        return self._with_queries
+
     @property
     def subqueries(self) -> Dict:
         """
@@ -694,12 +746,14 @@ class Parser:  # pylint: disable=R0902
         else:
             current_section.extend(column)
 
-    def _add_to_columns_aliases_subsection(self, token: SQLToken):
+    def _add_to_columns_aliases_subsection(
+        self, token: SQLToken, left_expand: bool = True
+    ) -> None:
         """
         Add alias to the section in which it appears in query
         """
         keyword = token.last_keyword_normalized
-        alias = token.left_expanded
+        alias = token.left_expanded if left_expand else token.value
         if (
             token.last_keyword_normalized in ["FROM", "WITH"]
             and token.find_nearest_token("(").is_with_columns_start
@@ -738,18 +792,42 @@ class Parser:  # pylint: disable=R0902
             alias_of = alias_token.left_expanded
         return alias_of
 
-    def _resolve_subquery_alias_to_column(self, subquery_alias) -> List[str]:
+    def _resolve_sub_queries(self, column: str) -> List[str]:
+        """
+        Resolve column names coming from sub queries and with queries to actual
+        column names as they appear in the query
+        """
+        column = self._resolve_nested_query(
+            subquery_alias=column,
+            nested_queries_names=self.subqueries_names,
+            nested_queries=self.subqueries,
+            already_parsed=self._subqueries_parsers,
+        )
+        if isinstance(column, str):
+            column = self._resolve_nested_query(
+                subquery_alias=column,
+                nested_queries_names=self.with_names,
+                nested_queries=self.with_queries,
+                already_parsed=self._with_parsers,
+            )
+        return column if isinstance(column, list) else [column]
+
+    @staticmethod
+    def _resolve_nested_query(
+        subquery_alias: str,
+        nested_queries_names: List[str],
+        nested_queries: Dict,
+        already_parsed: Dict,
+    ) -> Union[str, List[str]]:
         """
         Resolves subquery reference to the actual column in the subquery
         """
         parts = subquery_alias.split(".")
-        if len(parts) != 2 or parts[0] not in self.subqueries_names:
-            return [subquery_alias]
+        if len(parts) != 2 or parts[0] not in nested_queries_names:
+            return subquery_alias
         sub_query, column_name = parts[0], parts[-1]
-        sub_query_definition = self.subqueries.get(sub_query)
-        subparser = self._subqueries_parsers.setdefault(
-            sub_query, Parser(sub_query_definition)
-        )
+        sub_query_definition = nested_queries.get(sub_query)
+        subparser = already_parsed.setdefault(sub_query, Parser(sub_query_definition))
         # in subquery you cannot have more than one column with given name
         # so it either has to have an alias or only one column with given name exists
         if column_name in subparser.columns_aliases_names:
@@ -763,10 +841,31 @@ class Parser:  # pylint: disable=R0902
 
         if column_name == "*":
             return subparser.columns
-        column_index = [x.split(".")[-1] for x in subparser.columns].index(column_name)
+        try:
+            column_index = [x.split(".")[-1] for x in subparser.columns].index(
+                column_name
+            )
+        except ValueError as exc:
+            # handle case when column name is used but subquery select all by wildcard
+            if "*" in subparser.columns:
+                return column_name
+            raise exc  # pragma: no cover
         resolved_column = subparser.columns[column_index]
         return [resolved_column]
 
+    def _is_with_query_already_resolved(self, col_alias: str) -> bool:
+        """
+        Checks if columns comes from a with query that has columns defined
+        cause if it does that means that column name is an alias and is already
+        resolved in aliases.
+        """
+        parts = col_alias.split(".")
+        if len(parts) != 2 or parts[0] not in self.with_names:
+            return False
+        if self._with_queries_columns[parts[0]]:
+            return True
+        return False
+
     def _determine_opening_parenthesis_type(self, token: SQLToken):
         """
         Determines the type of left parenthesis in query
@@ -781,9 +880,9 @@ class Parser:  # pylint: disable=R0902
             token.is_column_definition_start = True
         elif token.previous_token.normalized == "AS":
             token.is_with_query_start = True
-        elif (
+        elif token.last_keyword_normalized == "TABLE" and (
             token.get_nth_previous(2).normalized == "TABLE"
-            and token.get_nth_previous(3).normalized == "CREATE"
+            or token.get_nth_previous(4).normalized == "TABLE"
         ):
             token.is_create_table_columns_declaration_start = True
         else:
